import logging


class HookHandler(logging.Handler):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def emit(self, record):
        self.fn(record)


class LogHook:
    def __init__(self, fn):
        self.handler = HookHandler(fn)
        self.logger = logging.getLogger()

    def __enter__(self):
        self.logger.addHandler(self.handler)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logger.removeHandler(self.handler)


def readable_time(duration):
    duration = int(duration.total_seconds())
    if duration == 0:
        return "<1s"
    elif duration < 60:
        return f"{duration}s"
    elif duration < 3600:
        mins = int(duration // 60)
        secs = duration % 60
        return f"{mins}m {secs}s"
    else:
        hours = int(duration // 3600)
        mins = int((duration % 3600) // 60)
        secs = duration % 60
        return f"{hours}h {mins}m {secs}s"



from __future__ import annotations

import logging
from collections import Counter, deque
from dataclasses import dataclass
from datetime import datetime
from logging import LogRecord

from serieux import TaggedSubclass

from .utils import LogHook


@dataclass
class Report:
    description: str
    reporters: list[TaggedSubclass[Reporter]]

    def __post_init__(self):
        self.start: datetime = None
        self.end: datetime = None
        self.statistics: Counter = Counter()
        self.errlogs: deque[LogRecord] = deque(maxlen=1000)
        self.exception: Exception = None
        self.message: str = None

    def set_message(self, message):
        self.message = message

    def on_log(self, lrec: LogRecord):
        self.statistics["log_" + lrec.levelname.lower()] += 1
        if lrec.levelno >= logging.ERROR:
            self.errlogs.append(lrec)

    def __enter__(self):
        self.start = datetime.now()
        self._loghook = LogHook(self.on_log)
        self._loghook.__enter__()
        for r in self.reporters:
            r.pre_report(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None:
            self.exception = exc_val
        self.end = datetime.now()
        for r in self.reporters:
            r.report(self)
        self._loghook.__exit__(exc_type, exc_val, exc_tb)
        return False


class Reporter:
    def log(self, markdown: str):
        raise NotImplementedError()

    def pre_report(self, report: Report):
        # OK if not implemented
        pass

    def report(self, report: Report):
        raise NotImplementedError()



import socket
import time
import traceback
from dataclasses import dataclass

from serieux.features.encrypt import Secret
from slack_sdk import WebClient

from .report import Report, Reporter
from .utils import readable_time


@dataclass(kw_only=True)
class SlackReporter(Reporter):
    token: Secret[str]
    channel: str
    show_logs: int = 15
    hostname: str = None

    def __post_init__(self):
        self.client = WebClient(token=self.token)

    def log(self, markdown: str = None, **kwargs):
        self.client.chat_postMessage(
            channel=self.channel,
            markdown_text=markdown,
            **kwargs,
        )

    def report(self, report: Report):
        duration_str = readable_time(report.end - report.start)

        success = not report.exception
        sprefix = "" if success else "un"

        icons = ""
        if n := report.statistics["log_warning"]:
            icons += f" ⚠️{n}"
        if n := report.statistics["log_error"]:
            icons += f" ❌{n}"

        blocks = []
        exception = report.exception
        if exception is not None:
            icons += f" (**raised** {type(exception).__name__})"
            tb_str = "".join(
                traceback.format_exception(
                    type(exception), exception, exception.__traceback__
                )
            )[-2900:]
            blocks.append(
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": f"*Exception traceback:*\n```{tb_str}```",
                    },
                }
            )
        if report.errlogs and (nerr := self.show_logs):
            error_lines = []
            for record in list(report.errlogs)[-nerr:]:
                ts = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(record.created))
                msg = record.getMessage()
                error_lines.append(f"{ts} [{record.levelname}] {record.name}: {msg}")
            error_block = "\n".join(error_lines)[-2900:]
            blocks.append(
                {
                    "type": "section",
                    "text": {
                        "type": "mrkdwn",
                        "text": f"*Last {nerr} errors logged:*\n```{error_block}```",
                    },
                }
            )

        hostname = self.hostname or socket.gethostname()
        summary = f"[{hostname}] **{report.description}** ran **{sprefix}successfully** in {duration_str}. {icons}"
        if report.message:
            summary += f"\n{report.message}"

        attachments = (
            [{"blocks": blocks, "fallback": "<error summary>"}] if blocks else []
        )

        self.client.chat_postMessage(
            channel=self.channel,
            markdown_text=summary,
            attachments=attachments,
            icon_emoji=(
                ":x:"
                if exception
                else (":warning:" if report.errlogs else ":white_check_mark:")
            ),
        )
